# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

from pynautobot.core.response import Record
from pynautobot.core.endpoint import DetailEndpoint


class IpAddresses(Record):
    def __str__(self):
        parent_record_string = super().__str__()
        return parent_record_string or str(self.address)


class Prefixes(Record):
    def __str__(self):
        parent_record_string = super().__str__()
        return parent_record_string or str(self.prefix)

    @property
    def available_ips(self):
        """Represents the ``available-ips`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing and creating IP addresses inside a prefix.

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> prefix.available_ips.list()
        [{u'vrf': None, u'family': 4, u'address': u'10.1.1.49/30'}...]

        To create a single IP:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> prefix.available_ips.create()
        {u'status': 1, u'description': u'', u'nat_inside': None...}


        To create multiple IPs:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> create = prefix.available_ips.create([{} for i in range(2)])
        >>> len(create)
        2
        """
        return DetailEndpoint(self, "available-ips", custom_return=IpAddresses)

    @property
    def available_prefixes(self):
        """Represents the ``available-prefixes`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing and creating prefixes inside a parent prefix.

        Very similar to :py:meth:`~pynautobot.ipam.Prefixes.available_ips`
        , except that dict (or list of dicts) passed to ``.create()``
        needs to have a ``prefix_length`` key/value specifed.

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        >>> prefix.available_prefixes.list()
        [{u'prefix': u'10.1.1.44/30', u'vrf': None, u'family': 4}]


        Creating a single child prefix:

        >>> prefix = nb.ipam.prefixes.get(1)
        >>> new_prefix = prefix.available_prefixes.create(
        ...    {'prefix_length': 29}
        ...)
        >>> new_prefix['prefix']
        u'10.1.1.56/29'

        """
        return DetailEndpoint(self, "available-prefixes", custom_return=Prefixes)



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

import copy
from collections import OrderedDict

import pynautobot.core.app
from requests.utils import urlparse
from pynautobot.core.query import Request
from pynautobot.core.util import Hashabledict


# List of fields that are lists but should be treated as sets.
LIST_AS_SET = ("tags", "tagged_vlans", "nat_outside")


def get_return(lookup, return_fields=None):
    """Returns simple representations for items passed to lookup.

    Args:
        lookup: The object or collection to retrieve representations for.
        return_fields (list, optional): A list of fields to reference when
            calling values on lookup.

    Returns:
        The simple representation of the object or collection.

    Note:
        Used to return a "simple" representation of objects and collections
        sent to it via lookup. Otherwise, we look to see if
        lookup is a "choices" field (dict with only 'id' and 'value')
        or a nested_return. Finally, we check if it's a Record, if
        so simply return a string. Order is important due to nested_return
        being self-referential.
    """

    for i in return_fields or ["id", "value", "nested_return"]:
        if isinstance(lookup, dict) and lookup.get(i):
            return lookup[i]
        else:
            if hasattr(lookup, i):
                # check if this is a "choices" field record
                # from a Nautobot 2.7 server.
                if sorted(dict(lookup)) == sorted(["id", "value", "label"]):
                    return getattr(lookup, "value")
                return getattr(lookup, i)

    if isinstance(lookup, Record):
        return str(lookup)
    else:
        return lookup


class JsonField(object):
    """Explicit field type for values that are not to be converted to a Record object."""

    _json_field = True


class Record(object):
    """Create Python objects from Nautobot API responses.

    Creates an object from a Nautobot response passed as `values`.
    Nested dicts that represent other endpoints are also turned
    into Record objects. All fields are then assigned to the
    object's attributes. If a missing attribute is requested
    (e.g., requesting a field that's only present on a full response on
    a Record made from a nested response), the pynautobot will make a
    request for the full object and return the requested value.

    Examples:
        Default representation of the object usually contains object's
        class, object's name, and its location in memory

        >>> x = nb.dcim.devices.get(1)
        >>> x
        <pynautobot.models.dcim.Devices ('test1-switch1') at 0x1953d821250>
        >>>

        Querying a string field.

        >>> x = nb.dcim.devices.get(1)
        >>> x.serial
        'ABC123'
        >>>

        Querying a field on a nested object.

        >>> x = nb.dcim.devices.get(1)
        >>> x.device_type.model
        'QFX5100-24Q'
        >>>

        Casting the object as a dictionary.

        >>> from pprint import pprint
        >>> pprint(dict(x))
        {'asset_tag': None,
         'cluster': None,
         'comments': '',
         'config_context': {},
         'created': '2018-04-01',
         'custom_fields': {},
         'device_role': {...},
         'device_type': {...},
         'display_name': 'test1-switch1',
         'face': {'label': 'Rear', 'value': 1},
         'id': 1,
         'name': 'test1-switch1',
         'parent_device': None,
         'platform': {...},
         'position': 1,
         'primary_ip': {...},
         'primary_ip4': {...},
         'primary_ip6': None,
         'rack': {...},
         'serial': 'ABC123',
         'site': {...},
         'status': {...},
         'tags': [],
         'tenant': None,
         'vc_position': None,
         'vc_priority': None,
         'virtual_chassis': None}

        Iterating over a Record object.

        >>> for i in x:
        ...     print(i)
        ...
        ('id', 1)
        ('name', 'test1-switch1')
        ('display_name', 'test1-switch1')
        >>>
    """

    url = None

    def __init__(self, values, api, endpoint):
        self.has_details = False
        self._full_cache = []
        self._init_cache = []
        self.api = api
        self.default_ret = Record
        self.endpoint = self._endpoint_from_url(values["url"]) if "url" in values else endpoint

        if values:
            self._parse_values(values)

    def __getattr__(self, k):
        """Default behavior for missing attributes.

        We'll call `full_details()` if we're asked for an attribute
        we don't have.

        Args:
            k (str): The name of the attribute.

        Returns:
            Any: The value of the requested attribute.

        Raises:
            AttributeError: If the attribute is not found.

        Note:
            In order to prevent non-explicit behavior, `k='keys'` is
            excluded because casting to dict() calls this attribute.
        """
        if self.url:
            if self.has_details is False and k != "keys":
                if self.full_details():
                    ret = getattr(self, k, None)
                    if ret or hasattr(self, k):
                        return ret

        raise AttributeError('object has no attribute "{}"'.format(k))

    def __iter__(self):
        for i in dict(self._init_cache):
            cur_attr = getattr(self, i)
            if isinstance(cur_attr, Record):
                yield i, dict(cur_attr)
            elif isinstance(cur_attr, list) and all(isinstance(i, Record) for i in cur_attr):
                yield i, [dict(x) for x in cur_attr]
            else:
                yield i, cur_attr

    def __getitem__(self, k):
        return dict(self)[k]

    def __str__(self):
        return str(getattr(self, "display", None) or getattr(self, "name", None) or getattr(self, "label", None) or "")

    def __repr__(self):
        return "<{}.{} ('{}') at {}>".format(self.__class__.__module__, self.__class__.__name__, self, hex(id(self)))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __key__(self):
        if hasattr(self, "id"):
            return (self.endpoint.name, self.id)
        else:
            return self.endpoint.name

    def __hash__(self):
        return hash(self.__key__())

    def __eq__(self, other):
        if isinstance(other, Record):
            return self.__key__() == other.__key__()
        return NotImplemented

    def _add_cache(self, item):
        key, value = item
        if key == "local_context_data":
            self._init_cache.append((key, copy.deepcopy(value)))
        else:
            self._init_cache.append((key, get_return(value)))

    def _parse_values(self, values):
        """Parses the values provided during initialization.

        Args:
            values (dict): A dictionary containing the values to be set as object attributes.

        Returns:
            None

        Note:
            This method sets object attributes using the values within the provided dictionary.
        """

        def list_parser(list_item):
            if isinstance(list_item, dict):
                return self.default_ret(list_item, self.api, self.endpoint)
            return list_item

        for k, v in values.items():
            if isinstance(v, dict):
                lookup = getattr(self.__class__, k, None)
                if k in ["custom_fields", "local_context_data"] or hasattr(lookup, "_json_field"):
                    self._add_cache((k, v.copy()))
                    setattr(self, k, v)
                    continue
                if lookup:
                    v = lookup(v, self.api, self.endpoint)
                else:
                    v = self.default_ret(v, self.api, self.endpoint)
                self._add_cache((k, v))

            elif isinstance(v, list):
                v = [list_parser(i) for i in v]
                to_cache = list(v)
                self._add_cache((k, to_cache))

            else:
                self._add_cache((k, v))
            setattr(self, k, v)

    def _endpoint_from_url(self, url):
        url_path = urlparse(url).path
        base_url_path_parts = urlparse(self.api.base_url).path.split("/")
        if len(base_url_path_parts) > 2:
            # There are some extra directories in the path, remove them from url
            extra_path = "/".join(base_url_path_parts[:-1])
            url_path = url_path[len(extra_path) :]
        split_url_path = url_path.split("/")
        if split_url_path[2] == "plugins":
            # Keep plugins in app path
            app = "/".join(split_url_path[2:4])
            name = "/".join(split_url_path[4:-2])
        else:
            app, name = split_url_path[2:4]
        return getattr(pynautobot.core.app.App(self.api, app), name)

    def full_details(self):
        """Queries the hyperlinked endpoint if 'url' is defined.

        This method will populate the attributes from the detail
        endpoint when it's called. Sets the class-level `has_details`
        attribute when it's called to prevent being called more
        than once.

        Returns:
            True
        """
        if self.url:
            req = Request(
                base=self.url,
                token=self.api.token,
                http_session=self.api.http_session,
                api_version=self.api.api_version,
            )
            self._parse_values(req.get())
            self.has_details = True
            return True
        return False

    def serialize(self, nested=False, init=False):
        """Serializes the object.

        Pulls all the attributes in an object and creates a dictionary
        that can be turned into the JSON format expected by Nautobot.

        Args:
            nested (bool): Whether to include nested objects in the serialization. Default is False.
            init (bool): Whether to include initialization attributes in the serialization. Default is False.

        Returns:
            dict: A dictionary representation of the serialized object.

        Note:
            Using this method to get a dictionary representation of the record
            is discouraged. It's probably better to cast to dict() instead.
            See the Record docstring for an example.
        """
        if nested:
            return get_return(self)

        if init:
            init_vals = dict(self._init_cache)

        ret = {}
        for i in dict(self):
            current_val = getattr(self, i) if not init else init_vals.get(i)
            if i in ["custom_fields", "constraints"]:  # just pass constraints as it is (a JSON string)
                ret[i] = current_val

            else:
                if isinstance(current_val, Record):
                    current_val = getattr(current_val, "serialize")(nested=True)

                if isinstance(current_val, list):
                    current_val = [v.id if isinstance(v, Record) else v for v in current_val]
                    if i in LIST_AS_SET and (
                        all([isinstance(v, str) for v in current_val]) or all([isinstance(v, int) for v in current_val])
                    ):
                        current_val = list(OrderedDict.fromkeys(current_val))
                ret[i] = current_val
        return ret

    def _diff(self):
        def fmt_dict(k, v):
            if isinstance(v, dict):
                return k, Hashabledict(v)
            if isinstance(v, list):
                return k, ",".join(map(str, v))
            return k, v

        current = Hashabledict({fmt_dict(k, v) for k, v in self.serialize().items()})
        init = Hashabledict({fmt_dict(k, v) for k, v in self.serialize(init=True).items()})
        return set([i[0] for i in set(current.items()) ^ set(init.items())])

    def updates(self):
        """Compiles changes for an existing object into a dictionary.

        Takes a diff between the object's current state and its state at initialization
        and returns them as a dictionary. Returns an empty dictionary if no changes
        have been made.

        Returns:
            dict: A dictionary containing the changes made to the object.

        Examples:
            >>> x = nb.dcim.devices.get(name='test1234')
            >>> x.serial
            ''
            >>> x.serial = '1234'
            >>> x.updates()
            {'serial': '1234'}
        """
        if self.id:
            diff = self._diff()
            if diff:
                serialized = self.serialize()
                return {i: serialized[i] for i in diff}
        return {}

    def save(self):
        """Saves changes to an existing object.

        Takes a diff between the object's current state and its state at initialization
        and sends them as a dictionary to Request.patch().

        Returns:
            bool: True if the PATCH request was successful.

        Examples:
            >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
            >>> x.serial
            ''
            >>> x.serial = '1234'
            >>> x.save()
            True
        """
        if self.id:
            diff = self._diff()
            if diff:
                serialized = self.serialize()
                req = Request(
                    key=self.id,
                    base=self.endpoint.url,
                    token=self.api.token,
                    http_session=self.api.http_session,
                    api_version=self.api.api_version,
                )
                if req.patch({i: serialized[i] for i in diff}):
                    return True

        return False

    def update(self, data):
        """Update an object with a dictionary.

        Accepts a dictionary and uses it to update the record and call save().
        For nested and choice fields you'd pass an int the same as if you were
        modifying the attribute and calling save().

        Args:
            data (dict): Dictionary containing the key-value pairs to update
                the record object with.

        Returns:
            bool: True if the PATCH request was successful.

        Examples:
            >>> x = nb.dcim.devices.get(1)
            >>> x.update({
            ...   "name": "test-switch2",
            ...   "serial": "ABC321",
            ... })
            True
        """

        for k, v in data.items():
            setattr(self, k, v)
        return self.save()

    def delete(self):
        """Deletes an existing object.

        Returns:
            bool: True if the DELETE operation was successful.

        Examples:
            >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
            >>> x.delete()
            True
        """
        req = Request(
            key=self.id,
            base=self.endpoint.url,
            token=self.api.token,
            http_session=self.api.http_session,
            api_version=self.api.api_version,
        )
        return True if req.delete() else False



"""GraphQL endpoint for making queries to the Nautobot GraphQL endpoint."""

from typing import Optional, Dict, Any


class GraphQLException(Exception):
    def __init__(self, graph_err):
        """GraphQL Exception handling.

        Args:
            graph_err (Exception): Exception from the request
        """
        self.json = graph_err.response.json()
        self.status_code = graph_err.response.status_code
        self.errors = graph_err.response.json()["errors"]
        self.url = graph_err.request.url
        self.body = graph_err.request.body
        self.reason = graph_err.response.reason

    def __str__(self):
        return str(self.errors)


class GraphQLRecord:
    def __init__(self, json: Dict[str, Any], status_code: int):
        """Initialization of class.

        Args:
            json (dict): The json response from making a graphql query.
            status_code (int): The HTTP status code from making the graphql query.
        """
        self.json = json
        self.status_code = status_code

    def __repr__(self):
        return f"GraphQLRecord(json={self.json}, status_code={self.status_code})"

    def __str__(self):
        return str(self.json)


class GraphQLQuery:
    def __init__(self, api):
        """Initialization of class.

        Args:
            api (pynautobot.api): pynatutobot Api class to make calls to the Nautobot API endpoints.
        """
        self.api = api
        # Add on top of the base_url the endpoint /graphql/ which will be the url used
        self.url = self.api.base_url + "/graphql/"

    def query(self, query: str, variables: Optional[Dict[str, Any]] = None) -> GraphQLRecord:
        """Runs query against Nautobot Graphql endpoint.

        Args:
            query (str): Query string to send to the API
            variables (dict): Dictionary of variables to use with the query string, defaults to None

        Raises:
            GraphQLException:
                - When the query string is invalid.
            TypeError:
                - When `query` passed in is not of type string.
                - When `variables` passed in is not a dictionary.
            Exception:
                - When unknown error is passed in, please open an issue so this can be addressed.

        Examples:
            >>> query = '''
            ... query {
            ...   sites {
            ...     id
            ...     name
            ...     region {
            ...       name
            ...     }
            ...   }
            ... }
            ... '''
            >>> nautobot.graphql.query(query=query)
            GraphQLRecord(json={'data': {'sites': [{'id': '2cfbc91e-361f-4129-9db6-bc21a6f88d38', 'name': 'ams', 'region': {'name': 'Netherlands'}}, {'id': '7ef021de-eb21-4403-8438-0e722c0f4b44', 'name': 'atl', 'region': {'name': 'United States'}}]}}, status_code=200)

            >>> query = '''
            ... query {
            ...   sites (name: "den") {
            ...     id
            ...     name
            ...     region {
            ...       name
            ...     }
            ...   }
            ... }
            ... '''
            >>> nautobot.graphql.query(query=query)
            GraphQLRecord(json={'data': {'sites': [{'id': '45399b54-47f9-4eec-86e3-47352e103b1b', 'name': 'den', 'region': {'name': 'United States'}}]}}, status_code=200)

            >>> variables = {"site_name": "den"}
            >>> query = '''
            ... query ($site_name:String!) {
            ...   sites (name: $site_name) {
            ...     id
            ...     name
            ...     region {
            ...       name
            ...     }
            ...   }
            ... }
            ... '''
            >>> nautobot.graphql.query(query=query, variables=variables)
            GraphQLRecord(json={'data': {'sites': [{'id': '45399b54-47f9-4eec-86e3-47352e103b1b', 'name': 'den', 'region': {'name': 'United States'}}]}}, status_code=200)

        Returns:
            GraphQLRecord: Response of the API call
        """
        # Check that the query data coming in is of a type of string
        if not isinstance(query, str):
            raise TypeError(f"Query should be of type string, not of type {type(query)}")

        # Check that variables coming in is a dictionary
        if variables is not None and not isinstance(variables, dict):
            raise TypeError(f"Variables should be of type dictionary, not of type {type(variables)}")

        payload = {"query": query, "variables": variables}

        response = self.api.http_session.post(self.url, json=payload, headers=self.api.headers)

        # Don't create an object with an error, raise an exception
        try:
            response.raise_for_status()
        except Exception as graph_err:
            if response.status_code == 400:
                raise GraphQLException(graph_err)
            else:
                raise graph_err

        return GraphQLRecord(json=response.json(), status_code=response.status_code)



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

import logging

from pynautobot.core.endpoint import Endpoint, JobsEndpoint
from pynautobot.core.query import Request
from pynautobot.models import circuits, dcim, extras, ipam, users, virtualization

logger = logging.getLogger(__name__)


class App(object):
    """Represents apps in Nautobot.

    Calls to attributes are returned as Endpoint objects.

    Returns:
        Endpoint: Matching requested attribute.

    Raises:
        RequestError: If requested endpoint doesn't exist.
    """

    models = {
        "dcim": dcim,
        "ipam": ipam,
        "circuits": circuits,
        "virtualization": virtualization,
        "extras": extras,
        "users": users,
    }

    def __init__(self, api, name):
        self.api = api
        self.name = name
        self._choices = None
        self._setmodel()

    def _setmodel(self):
        self.model = App.models[self.name] if self.name in App.models else None

    def __getstate__(self):
        return {"api": self.api, "name": self.name, "_choices": self._choices}

    def __setstate__(self, d):
        self.__dict__.update(d)
        self._setmodel()

    def __getattr__(self, name):
        if name == "jobs":
            return JobsEndpoint(self.api, self, name, model=self.model)
        return Endpoint(self.api, self, name, model=self.model)

    def choices(self):
        """Returns _choices response from App.

        Returns:
            Raw response from Nautobot's _choices endpoint.
        """
        if self._choices:
            return self._choices

        self._choices = Request(
            base="{}/{}/_choices/".format(self.api.base_url, self.name),
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()

        return self._choices

    def get_custom_fields(self):
        """Returns custom-fields response from app.

        Returns:
            Raw response from Nautobot's custom-fields endpoint.

        Raises:
            RequestError: If called for an invalid endpoint.

        Examples:
            >>> nb.extras.get_custom_fields()
            [
                {
                    "id": "5b39ba88-e5ab-4be2-89f5-5a016473b53c",
                    "display": "Test custom field",
                    "url": "http://localhost:8000/api/extras/custom-fields/5b39ba88-e5ab-4be2-89f5-5a016473b53c/",
                    "content_types": ["dcim.rack"],
                    "type": {"value": "integer", "label": "Integer"},
                    "label": "Test custom field",
                    "name": "test_custom_field",
                    "slug": "test_custom_field",
                    "description": "",
                    "required": False,
                    "filter_logic": {"value": "loose", "label": "Loose"},
                    "default": None,
                    "weight": 100,
                    "validation_minimum": None,
                    "validation_maximum": None,
                    "validation_regex": "",
                    "created": "2023-04-15",
                    "last_updated": "2023-04-15T17:45:11.839431Z",
                    "notes_url": "http://localhost:8000/api/extras/custom-fields/5b39ba88-e5ab-4be2-89f5-5a016473b53c/notes/",
                },
            ]
        """
        return Request(
            base=f"{self.api.base_url}/{self.name}/custom-fields/",
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()

    def get_custom_field_choices(self):
        """Returns custom-field-choices response from app.

        Returns:
            Raw response from Nautobot's custom-field-choices endpoint.

        Raises:
            RequestError: If called for an invalid endpoint.

        Examples:
            >>> nb.extras.get_custom_field_choices()
            [
                {
                    "id": "5b39ba88-e5ab-4be2-89f5-5a016473b53c",
                    "display": "First option",
                    "url": "http://localhost:8000/api/extras/custom-field-choices/5b39ba88-e5ab-4be2-89f5-5a016473b53c/",
                    "field": {
                        "display": "Test custom field 2",
                        "id": "5b39ba88-e5ab-4be2-89f5-5a016473b53c",
                        "url": "http://localhost:8000/api/extras/custom-fields/5b39ba88-e5ab-4be2-89f5-5a016473b53c/",
                        "name": "test_custom_field_2"
                    },
                    "value": "First option",
                    "weight": 100,
                    "created": "2023-04-15",
                    "last_updated": "2023-04-15T18:11:57.163237Z"
                },
            ]
        """
        return Request(
            base=f"{self.api.base_url}/{self.name}/custom-field-choices/",
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()

    def config(self):
        """Returns config response from app.

        Returns:
            dict: Raw response from Nautobot's config endpoint.

        Raises:
            RequestError: If called for an invalid endpoint.

        Example:
            >>> pprint.pprint(nb.users.config())
            {'tables': {'DeviceTable': {'columns': ['name',
                                                    'status',
                                                    'tenant',
                                                    'device_role',
                                                    'site',
                                                    'primary_ip',
                                                    'tags']}}}
        """

        config = Request(
            base="{}/{}/config/".format(
                self.api.base_url,
                self.name,
            ),
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()
        return config


class PluginsApp(object):
    """Add plugins to the URL path.

    Basically, valid plugins API could be handled by the same App class,
    but you need to add "plugins" to the request URL path.

    Returns:
        App: With "plugins" added to the path.
    """

    def __init__(self, api):
        self.api = api

    def __getattr__(self, name):
        return App(self.api, f"plugins/{name.replace('_', '-')}")

    def installed_plugins(self):
        """Returns raw response with installed plugins.

        Returns:
            Raw response from Nautobot's installed plugins.

        Examples:
            >>> nb.plugins.installed_plugins()
            [{
                'name': 'test_plugin',
                'package': 'test_plugin',
                'author': 'Dmitry',
                'description': 'Nautobot test plugin',
                'verison': '0.10'
            }]
        """
        installed_plugins = Request(
            base="{}/plugins/installed-plugins".format(
                self.api.base_url,
            ),
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()
        return installed_plugins



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

from packaging import version
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from pynautobot.core.query import Request
from pynautobot.core.app import App, PluginsApp
from pynautobot.core.graphql import GraphQLQuery


class Api(object):
    """The `Api` object is the primary entry point for interacting with a Nautobot
    instance using pynautobot.

    Args:
        url (str): The base URL of the Nautobot instance you want to connect to.
        token (str): Your Nautobot authentication token.
        threading (bool, optional): Enable threading for `.all()` and `.filter()`
            requests. Defaults to `False`.
        max_workers (int, optional): The maximum number of worker threads to use
            for `.all()` and `.filter()` requests. Defaults to the number of CPU cores.
        api_version (str, optional): Override the default Nautobot REST API version
            used for all requests.
        retries (int, optional): The number of retries for HTTP status codes
            429, 500, 502, 503, and 504. Defaults to 0 (no retries).
        verify (bool, optional): Whether to verify SSL certificates. Defaults to `True`.

    Attributes:
        dcim: An instance of the `App` class providing access to DCIM endpoints.
        ipam: An instance of the `App` class providing access to IPAM endpoints.
        circuits: An instance of the `App` class providing access to Circuits endpoints.
        tenancy: An instance of the `App` class providing access to Tenancy endpoints.
        extras: An instance of the `App` class providing access to Extras endpoints.
        virtualization: An instance of the `App` class providing access to Virtualization endpoints.
        users: An instance of the `App` class providing access to User endpoints.
        http_session (requests.Session): The underlying HTTP session object used for
            making requests to Nautobot. You can override the default session with your
            own to control HTTP behavior such as SSL verification, custom headers,
            retries, and timeouts. See the documentation on custom sessions
            for more information.

    Raises:
        AttributeError: If an invalid application name is provided.

    Examples:
        >>> import pynautobot
        >>> nb = pynautobot.api(
        ...     'http://localhost:8000',
        ...     token='d6f4e314a5b5fefd164995169f28ae32d987704f'
        ... )
        >>> nb.dcim.devices.all()
    """

    def __init__(
        self,
        url,
        token=None,
        threading=False,
        max_workers=4,
        api_version=None,
        retries=0,
        verify=True,
    ):
        base_url = "{}/api".format(url if url[-1] != "/" else url[:-1])
        self.token = token
        self.headers = {"Authorization": f"Token {self.token}"}
        self.base_url = base_url
        self.http_session = requests.Session()
        self.http_session.verify = verify
        if retries:
            _adapter = HTTPAdapter(
                max_retries=Retry(
                    total=retries,
                    backoff_factor=1,
                    allowed_methods=None,
                    status_forcelist=[429, 500, 502, 503, 504],
                )
            )
            self.http_session.mount("http://", _adapter)
            self.http_session.mount("https://", _adapter)
        self.threading = threading
        self.max_workers = max_workers
        self.api_version = api_version

        self.dcim = App(self, "dcim")
        self.ipam = App(self, "ipam")
        self.circuits = App(self, "circuits")
        self.tenancy = App(self, "tenancy")
        self.extras = App(self, "extras")
        self.virtualization = App(self, "virtualization")
        self.users = App(self, "users")
        self.plugins = PluginsApp(self)
        self.graphql = GraphQLQuery(self)
        self._validate_version()

    def _validate_version(self):
        """Validate API version if eq or ge than 2.0 raise an error."""
        api_version = self.version
        if api_version.replace(".", "").isnumeric() and version.parse(api_version) < version.parse("2.0"):
            raise ValueError("Nautobot version 1 detected, please downgrade pynautobot to version 1.x")

    @property
    def version(self):
        """Retrieves the version of the Nautobot REST API that the connected instance is using.

        This method can be helpful for checking API compatibility and determining
        if specific features or syntaxes are available.

        Returns:
            str: The Nautobot API version string.

        Raises:
            requests.exceptions.RequestException: If there is an error fetching the
                version information from Nautobot.

        Examples:
            >>> import pynautobot
            >>> nb = pynautobot.api(
            ...     'http://localhost:8000',
            ...     token='d6f4e314a5b5fefd164995169f28ae32d987704f'
            ... )
            >>> nb.version
            '1.0'
        """

        return Request(
            base=self.base_url,
            http_session=self.http_session,
            api_version=self.api_version,
            token=self.token,
        ).get_version()

    def openapi(self):
        """Retrieves the OpenAPI specification (OAS) document for the connected Nautobot instance.

        The OpenAPI specification provides a machine-readable description of the Nautobot REST API,
        including available endpoints, parameters, request and response formats, and more.

        This can be useful for tools like code generation or API clients in other languages.

        Returns:
            dict: The OpenAPI specification document as a Python dictionary.

        Raises:
            requests.exceptions.RequestException: If there is an error fetching the
                OpenAPI spec from Nautobot.

        Examples:
            >>> import pynautobot
            >>> nb = pynautobot.api(
            ...     'http://localhost:8000',
            ...     token='d6f4e314a5b5fefd164995169f28ae32d987704f'
            ... )
            >>> nb.openapi()
            {...}
        """

        return Request(
            base=self.base_url,
            http_session=self.http_session,
            api_version=self.api_version,
            token=self.token,
        ).get_openapi()

    def status(self):
        """Retrieves status information about the connected Nautobot instance.

        This method provides various details about the Nautobot instance, including:

        * Django version
        * Installed apps and their versions
        * Nautobot version
        * Active plugins (if any)
        * Python version
        * Number of running RQ workers (if applicable)

        **Availability:** Requires Nautobot version 2.10.0 or newer.

        Returns:
            dict: A dictionary containing the status information as returned by Nautobot.

        Raises:
            pynautobot.exceptions.RequestError: If the request to Nautobot fails.

        Examples:
            >>> import pprint
            >>> nb = pynautobot.api(
            ...     'http://localhost:8000',
            ...     token='d6f4e314a5b5fefd164995169f28ae32d987704f'
            ... )
            >>> pprint.pprint(nb.status())
            {'django-version': '3.1.3',
            'installed-apps': {...},
            'nautobot-version': '1.0.0',
            'plugins': {},
            'python-version': '3.7.3',
            'rq-workers-running': 1}
        """
        return Request(
            base=self.base_url,
            token=self.token,
            http_session=self.http_session,
            api_version=self.api_version,
        ).get_status()



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

try:
    import concurrent.futures as cf
except ImportError:
    pass
import json
import requests


def calc_pages(limit, count):
    """Calculate number of pages required for full results set."""
    return int(count / limit) + (limit % count > 0)


class RequestError(Exception):
    """Basic Request Exception

    More detailed exception that returns the original requests object
    for inspection. Along with some attributes with specific details
    from the requests object. If return is JSON we decode and add it
    to the message.

    Examples:
        >>> try:
        ...     nb.dcim.devices.create(name="destined-for-failure")
        ... except pynautobot.RequestError as e:
        ...     print(e.error)
    """

    def __init__(self, message):
        req = message

        if req.status_code == 404:
            message = "The requested url: {} could not be found.".format(req.url)
        else:
            try:
                message = "The request failed with code {} {}: {}".format(req.status_code, req.reason, req.json())
            except ValueError:
                message = (
                    "The request failed with code {} {} but more specific "
                    "details were not returned in json. Check the Nautobot Logs "
                    "or investigate this exception's error attribute.".format(req.status_code, req.reason)
                )

        super(RequestError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = req.text


class RequestErrorFromException(Exception):
    """RequestErrorFromException is raised from exception."""


class AllocationError(Exception):
    """Allocation Exception

    Used with available-ips/available-prefixes when there is no
    room for allocation and Nautobot returns 204 No Content.
    """

    def __init__(self, message):
        req = message

        message = "The requested allocation could not be fulfilled."

        super(AllocationError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = message


class ContentError(Exception):
    """Content Exception

    If the API URL does not point to a valid Nautobot API, the server may
    return a valid response code, but the content is not json. This
    exception is raised in those cases.
    """

    def __init__(self, message):
        req = message

        message = "The server returned invalid (non-json) data. Maybe not " "a Nautobot server?"

        super(ContentError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = message


class Request(object):
    """Creates requests to the Nautobot API.

    Responsible for building the URL and making the HTTP(S) requests to
    Nautobot's API.

    Args:
        base (str): Base URL passed in api() instantiation.
        filters (dict, optional): Contains key/value pairs that
            correlate to the filters a given endpoint accepts.
            In (e.g. /api/dcim/devices/?name='test'), 'name': 'test'
            would be in the filters dict.
        max_workers (int, optional): Set the maximum workers for threading in ``.all()``
            and ``.filter()`` requests.
    """

    def __init__(
        self,
        base,
        http_session,
        filters=None,
        key=None,
        token=None,
        threading=False,
        max_workers=4,
        api_version=None,
    ):
        """Instantiates a new Request object

        Args:
            base (string): Base URL passed in api() instantiation.
            filters (dict, optional): contains key/value pairs that
                correlate to the filters a given endpoint accepts.
                In (e.g. /api/dcim/devices/?name='test') 'name': 'test'
                would be in the filters dict.
            key (int, optional): database id of the item being queried.
            api_version (str, optional): Set to override the default Nautobot REST API Version.
        """
        self.base = self.normalize_url(base)
        self.filters = filters
        self.key = key
        self.token = token
        self.http_session = http_session
        self.url = self.base if not key else "{}{}/".format(self.base, key)
        self.threading = threading
        self.max_workers = max_workers
        self.api_version = api_version

    def get_openapi(self):
        """Gets the OpenAPI Spec"""
        headers = {
            "Content-Type": "application/json;",
            "Authorization": f"Token {self.token}",
        }

        if self.api_version:
            headers["accept"] = f"application/json; version={self.api_version}"

        try:
            req = self.http_session.get(
                "{}docs/?format=openapi".format(self.normalize_url(self.base)),
                headers=headers,
            )
        except requests.exceptions.RetryError as error:
            raise RequestErrorFromException from error

        if req.ok:
            return req.json()
        else:
            raise RequestError(req)

    def get_version(self):
        """Gets the API version of Nautobot.

        Issues a GET request to the base URL to read the API version from the
        response headers.

        Raises:
            RequestError: If req.ok returns false.

        Returns:
            str: Version number as a string. Empty string if version is not present in the headers.
        """
        headers = {
            "Content-Type": "application/json;",
            "Authorization": f"Token {self.token}",
        }
        if self.api_version:
            headers["accept"] = f"application/json; version={self.api_version}"

        try:
            req = self.http_session.get(
                self.normalize_url(self.base),
                headers=headers,
            )
        except requests.exceptions.RetryError as error:
            raise RequestErrorFromException from error

        if req.ok:
            return req.headers.get("API-Version", "")
        else:
            raise RequestError(req)

    def get_status(self):
        """Gets the status from /api/status/ endpoint in Nautobot.

        Returns:
            dict: Dictionary as returned by Nautobot.

        Raises:
            RequestError: If request is not successful.
        """

        headers = {
            "Content-Type": "application/json;",
            "Authorization": f"Token {self.token}",
        }
        if self.token:
            headers["authorization"] = "Token {}".format(self.token)

        if self.api_version:
            headers["accept"] = f"application/json; version={self.api_version}"

        try:
            req = self.http_session.get(
                "{}status/".format(self.normalize_url(self.base)),
                headers=headers,
            )
        except requests.exceptions.RetryError as error:
            raise RequestErrorFromException from error

        if req.ok:
            return req.json()
        else:
            raise RequestError(req)

    def normalize_url(self, url):
        """Builds a url for POST actions."""
        if url[-1] != "/":
            return "{}/".format(url)

        return url

    def _make_call(self, verb="get", url_override=None, add_params=None, data=None):
        if verb in ("post", "put") or (verb in ("delete") and data):
            headers = {"Content-Type": "application/json;"}
        else:
            headers = {"accept": "application/json;"}

        if self.token:
            headers["authorization"] = "Token {}".format(self.token)

        if self.api_version:
            headers["accept"] = f"application/json; version={self.api_version}"

        params = {}
        if not url_override:
            if self.filters:
                params.update(self.filters)
            if add_params:
                params.update(add_params)

        try:
            req = getattr(self.http_session, verb)(url_override or self.url, headers=headers, params=params, json=data)
        except requests.exceptions.RetryError as error:
            raise RequestErrorFromException from error

        if req.status_code == 204 and verb == "post":
            raise AllocationError(req)
        if verb == "delete":
            if req.ok:
                return True
            else:
                raise RequestError(req)
        elif req.ok:
            try:
                return req.json()
            except json.JSONDecodeError:
                raise ContentError(req)
        else:
            raise RequestError(req)

    def concurrent_get(self, ret, page_size, page_offsets):
        futures_to_results = []
        with cf.ThreadPoolExecutor(max_workers=self.max_workers) as pool:
            for offset in page_offsets:
                new_params = {"offset": offset, "limit": page_size}
                futures_to_results.append(pool.submit(self._make_call, add_params=new_params))

            for future in cf.as_completed(futures_to_results):
                result = future.result()
                ret.extend(result["results"])

    def get(self, add_params=None):
        """Makes a GET request.

        Makes a GET request to Nautobot's API, and automatically recurses
        any paginated results.

        Raises:
            RequestError: If req.ok returns false.
            ContentError: If response is not JSON.

        Returns:
            List[Response]: List of `Response` objects returned from the
                endpoint.
        """

        def req_all():
            req = self._make_call(add_params=add_params)
            if isinstance(req, dict) and req.get("results") is not None:
                ret = req["results"]
                first_run = True
                while req["next"]:
                    # Not worrying about making sure add_params kwargs is
                    # passed in here because results from detail routes aren't
                    # paginated, thus far.
                    if first_run:
                        req = self._make_call(add_params={"limit": req["count"], "offset": len(req["results"])})
                    else:
                        req = self._make_call(url_override=req["next"])
                    first_run = False
                    ret.extend(req["results"])
                return ret
            else:
                return req

        def req_all_threaded(add_params):
            if add_params is None:
                # Limit must be 0 to discover the max page size
                add_params = {"limit": 0}
            req = self._make_call(add_params=add_params)
            if isinstance(req, dict) and req.get("results") is not None:
                ret = req["results"]
                if req.get("next"):
                    page_size = len(req["results"])
                    pages = calc_pages(page_size, req["count"])
                    page_offsets = [increment * page_size for increment in range(1, pages)]
                    if pages == 1:
                        req = self._make_call(url_override=req.get("next"))
                        ret.extend(req["results"])
                    else:
                        self.concurrent_get(ret, page_size, page_offsets)

                return ret
            else:
                return req

        if self.threading:
            return req_all_threaded(add_params)

        return req_all()

    def put(self, data: dict) -> dict:
        """Makes a PUT request to the Nautobot API.

        Args:
            data: (dict) The data to be sent in the PUT request body. It will be
                serialized to JSON before sending.

        Raises:
            RequestError: If the request fails (i.e., req.ok returns False).
            ContentError: If the response cannot be deserialized from JSON.

        Returns:
            dict: The deserialized JSON response from the Nautobot API.
        """
        return self._make_call(verb="put", data=data)

    def post(self, data: dict) -> dict:
        """Makes a POST request to the Nautobot API.

        Args:
            data: (dict) The data to be sent in the POST request body. It will be
                serialized to JSON before sending.

        Raises:
            RequestError: If the request fails (i.e., req.ok returns False).
            AllocationError: If the status code is 204 (No Content), indicating
                no available resources for allocation (e.g., available IPs or prefixes).
            ContentError: If the response cannot be deserialized from JSON.

        Returns:
            dict: The deserialized JSON response from the Nautobot API.
        """
        return self._make_call(verb="post", data=data)

    def delete(self, data=None):
        """Makes DELETE request.

        Makes a DELETE request to Nautobot's API.

        Args:
            data (list): Contains a dict that will be turned into a json object
                and sent to the API.

        Returns:
            bool: True if successful.

        Raises:
            RequestError: If req.ok doesn't return True.
        """
        return self._make_call(verb="delete", data=data)

    def patch(self, data: dict) -> dict:
        """Makes a PATCH request to the Nautobot API.

        Args:
            data: (dict) The data to be sent in the PATCH request body. It will be
                serialized to JSON before sending.

        Raises:
            RequestError: If the request fails (i.e., req.ok returns False).
            ContentError: If the response cannot be deserialized from JSON.

        Returns:
            dict: The deserialized JSON response from the Nautobot API.
        """
        return self._make_call(verb="patch", data=data)

    def options(self) -> dict:
        """Retrieves allowed HTTP methods for a Nautobot API endpoint.

        Makes an OPTIONS request to determine the allowed HTTP methods
        supported by a specific Nautobot API endpoint.

        Raises:
            RequestError: If the request fails (i.e., req.ok returns False).
            ContentError: If the response cannot be deserialized from JSON.

        Returns:
            dict: The deserialized JSON response from the Nautobot API,
                containing information about allowed methods.
        """

        return self._make_call(verb="options")

    def get_count(self, *args, **kwargs) -> int:
        """Retrieves the number of objects matching a query in the Nautobot API.

        Makes a GET request to the specified endpoint with a limited response
        (limit=1) and extracts the "count" field from the response to determine
        the total number of objects that would be returned by a full query
        with the same parameters.

        Args:
            *args: Additional positional arguments to be passed to the
                Nautobot API endpoint.
            **kwargs: Additional keyword arguments to be passed to the
                Nautobot API endpoint.

        Raises:
            RequestError: If the request fails (i.e., req.ok returns False).
            ContentError: If the response cannot be deserialized from JSON.

        Returns:
            int: The total number of objects that would match the provided query.
        """

        return self._make_call(add_params={"limit": 1})["count"]



"""
This file has been modified by NetworktoCode, LLC.
"""

from importlib.metadata import PackageNotFoundError, version

from pynautobot.core.api import Api as api
from pynautobot.core.query import AllocationError, ContentError, RequestError

__all__ = ["RequestError", "AllocationError", "ContentError", "api", "__version__"]


try:
    __version__ = version(__package__)
except PackageNotFoundError:
    pass



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

from typing import List, Dict, Any
from uuid import UUID
from pynautobot.core.query import Request, RequestError
from pynautobot.core.response import Record

RESERVED_KWARGS = ("pk", "limit", "offset")


def response_loader(req, return_obj, endpoint):
    if isinstance(req, list):
        return [return_obj(i, endpoint.api, endpoint) for i in req]
    return return_obj(req, endpoint.api, endpoint)


class Endpoint(object):
    """Represent actions available on endpoints in the Nautobot API.

    Takes `name` and `app` passed from `App()` and builds the correct
    URL to make queries to and the proper Response object to return
    results in.

    Args:
        api (object): Takes `Api` created at instantiation.
        app (object): Takes `App`.
        name (str): Name of endpoint passed to `App()`.
        model (object, optional): Custom model for given app.

    Note:
        In order to call Nautobot endpoints with dashes in their
        names you should convert the dash to an underscore.
        (E.g. querying the ip-addresses endpoint is done with
        `nb.ipam.ip_addresses.all()`.)
    """

    def __init__(self, api, app, name, model=None):
        self.return_obj = self._lookup_ret_obj(name, model)
        self.name = name.replace("_", "-")
        self.api = api
        self.base_url = api.base_url
        self.token = api.token
        self.url = "{base_url}/{app}/{endpoint}".format(
            base_url=self.base_url,
            app=app.name,
            endpoint=self.name,
        )
        self._choices = None

    def _lookup_ret_obj(self, name, model):
        """Loads unique Response objects.

        This method loads a unique response object for an endpoint if
        it exists. Otherwise, it returns a generic `Record` object.

        Args:
            name (str): Endpoint name.
            model (obj): The application model that contains unique Record objects.

        Returns:
            Record: Unique response object if exists, otherwise a generic `Record` object.
        """
        if model:
            name = name.title().replace("_", "").replace("-", "")
            ret = getattr(model, name, Record)
        else:
            ret = Record
        return ret

    def all(self, api_version=None):
        """Queries the 'ListView' of a given endpoint.

        Returns all objects from an endpoint.

        Args:
            api_version (str, optional): Override default or globally-set Nautobot REST API
                version for this single request.

        Returns:
            list: List of :py:class:`.Record` objects.

        Examples:
            >>> nb.dcim.devices.all()
            [test1-a3-oobsw2, test1-a3-oobsw3, test1-a3-oobsw4]
        """

        api_version = api_version or self.api.api_version
        req = Request(
            base="{}/".format(self.url),
            token=self.token,
            http_session=self.api.http_session,
            threading=self.api.threading,
            max_workers=self.api.max_workers,
            api_version=api_version,
        )

        return response_loader(req.get(), self.return_obj, self)

    def get(self, *args, **kwargs):
        """Queries the DetailsView of a given endpoint.

        Args:
            key (int, optional): ID for the item to be retrieved.
            **kwargs (str, optional): Accepts the same keyword args as filter().
                Any search argument the endpoint accepts can be added as a keyword arg.
            api_version (str, optional): Override default or globally-set Nautobot REST API
                version for this single request.

        Returns:
            Union[Record, None]: A single :py:class:`.Record` object or None.

        Raises:
            ValueError: If kwarg search returns more than one value.

        Examples:
            Referencing with a kwarg that only returns one value.
            >>> nb.dcim.devices.get(name='test1-a3-tor1b')
            test1-a3-tor1b

            Referencing with an id.
            >>> nb.dcim.devices.get(1)
            test1-edge1
        """

        try:
            key = args[0]
        except IndexError:
            key = None

        is_api_version = kwargs.pop("api_version") if kwargs.get("api_version") else None
        api_version = is_api_version or self.api.api_version

        if not key:
            filter_lookup = self.filter(**kwargs)
            if filter_lookup:
                if len(filter_lookup) > 1:
                    raise ValueError(
                        "get() returned more than one result. "
                        "Check that the kwarg(s) passed are valid for this "
                        "endpoint or use filter() or all() instead."
                    )
                else:
                    return filter_lookup[0]
            return None

        req = Request(
            key=key,
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
            api_version=api_version,
        )

        try:
            resp = req.get()
        except RequestError as e:
            if e.req.status_code == 404:
                return None
            else:
                raise e

        return response_loader(resp, self.return_obj, self)

    def filter(self, *args, api_version=None, **kwargs):
        """Queries the 'ListView' of a given endpoint.

        Takes named arguments that match the usable filters on a
        given endpoint. If an argument is passed then it's used as a
        freeform search argument if the endpoint supports it.

        Args:
            *args (str, optional): Freeform search string that's
                accepted on given endpoint.
            **kwargs (str, optional): Any search argument the
                endpoint accepts can be added as a keyword arg.
            api_version (str, optional): Override default or globally-set
                Nautobot REST API version for this single request.

        Returns:
            list: A list of :py:class:`.Record` objects.

        Examples:
            To return a list of objects matching a named argument filter.
            >>> nb.dcim.devices.filter(role='leaf-switch')
            [test1-a3-tor1b, test1-a3-tor1c, test1-a3-tor1d, test1-a3-tor2a]

            Using a freeform query along with a named argument.
            >>> nb.dcim.devices.filter('a3', role='leaf-switch')
            [test1-a3-tor1b, test1-a3-tor1c, test1-a3-tor1d, test1-a3-tor2a]

            Chaining multiple named arguments.
            >>> nb.dcim.devices.filter(role='leaf-switch', status=True)
            [test1-leaf2]

            Passing a list as a named argument adds multiple filters of the
            same value.
            >>> nb.dcim.devices.filter(role=['leaf-switch', 'spine-switch'])
            [test1-a3-spine1, test1-a3-spine2, test1-a3-leaf1]
        """

        if args:
            kwargs.update({"q": args[0]})

        if not kwargs:
            raise ValueError("filter must be passed kwargs. Perhaps use all() instead.")
        if any(i in RESERVED_KWARGS for i in kwargs):
            raise ValueError("A reserved {} kwarg was passed. Please remove it " "try again.".format(RESERVED_KWARGS))

        api_version = api_version or self.api.api_version
        req = Request(
            filters=kwargs,
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
            threading=self.api.threading,
            api_version=api_version,
        )

        return response_loader(req.get(), self.return_obj, self)

    def create(self, *args, api_version=None, **kwargs):
        """Creates an object on an endpoint.

        Allows for the creation of new objects on an endpoint. Named
        arguments are converted to JSON properties, and a single object
        is created. Nautobot's bulk creation capabilities can be used by
        passing a list of dictionaries as the first argument.

        Note:
            Any positional arguments will supersede named ones.

        Args:
            *args (list): A list of dictionaries containing the
                properties of the objects to be created.
            **kwargs (str): Key/value strings representing
                properties on a JSON object.
            api_version (str, optional): Override default or globally-set
                Nautobot REST API version for this single request.

        Returns:
            Union[Record, List[Record]]: A list or single :py:class:`.Record` object depending
                on whether a bulk creation was requested.

        Examples:
            Creating an object on the `devices` endpoint you can look up a
            device_role's name with:
            >>> nautobot.dcim.devices.create(
            ...    name='test',
            ...    device_role=1,
            ... )

            Use bulk creation by passing a list of dictionaries:
            >>> nb.dcim.devices.create([
            ...     {
            ...         "name": "test1-core3",
            ...         "device_role": 3,
            ...         "site": 1,
            ...         "device_type": 1,
            ...         "status": 1
            ...     },
            ...     {
            ...         "name": "test1-core4",
            ...         "device_role": 3,
            ...         "site": 1,
            ...         "device_type": 1,
            ...         "status": 1
            ...     }
            ... ])
        """

        api_version = api_version or self.api.api_version

        req = Request(
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
            api_version=api_version,
        ).post(args[0] if args else kwargs)

        return response_loader(req, self.return_obj, self)

    def update(self, *args, **kwargs):
        """Update a single resource with a dictionary or bulk update a list of objects.

        Allows for bulk updating of existing objects on an endpoint.
        Objects is a list which contain either JSON/dicts or Record
        derived objects, which contain the updates to apply.
        If JSON/dicts are used, then the id of the object *must* be
        included.

        Args:
            *args (list, optional): A list of dicts or a list of Record.
            **kwargs (str, optional): See Below.

        Keyword Arguments:
            * *id* (``string``) -- Identifier of the object being updated.
            * *data* (``dict``) -- Key/value pairs to update the record object with.

        Returns:
            Union[Record, List[Record]]: A list or single :py:class:`.Record` object depending
                on whether a bulk update was requested.

        Examples:
            Accepts the id of the object that needs to be updated as well as a
            dictionary of k/v pairs used to update an object.
            >>> nb.dcim.devices.update(id="0238a4e3-66f2-455a-831f-5f177215de0f", data={
            ...     "name": "test",
            ...     "serial": "1234",
            ...     "location": "9b1f53c7-89fa-4fb2-a89a-b97364fef50c",
            ... })

            Use bulk update by passing a list of dicts:
            >>> devices = nb.dcim.devices.update([
            ...    {'id': "db8770c4-61e5-4999-8372-e7fa576a4f65", 'name': 'test'},
            ...    {'id': "e9b5f2e0-4f20-41ad-9179-90a4987f743e", 'name': 'test2'},
            ... ])

            Use bulk update by passing a list of Records:
            >>> devices = list(nb.dcim.devices.filter())
            >>> devices
            [Device1, Device2, Device3]
            >>> for d in devices:
            ...     d.name = d.name+'-test'
            ...
            >>> nb.dcim.devices.update(devices)
        """
        if not args and not kwargs:
            raise ValueError("You must provide either a UUID and data dict or a list of objects to update")
        uuid = kwargs.get("id", "")
        data = kwargs.get("data", {})
        if data and not uuid:
            uuid = args[0]
        if len(args) == 2:
            uuid, data = args

        if not any([uuid, data]):
            return self.bulk_update(args[0])

        req = Request(
            key=uuid,
            base=self.url,
            token=self.api.token,
            http_session=self.api.http_session,
            api_version=self.api.api_version,
        )
        if req.patch(data):
            return True
        return False

    def bulk_update(self, objects: List[Dict[str, Any]]):
        """This method is called from the update() method if a bulk
        update is detected.

        Allows for bulk updating of existing objects on an endpoint.
        Objects is a list which contain either JSON/dicts or Record
        derived objects, which contain the updates to apply.
        If JSON/dicts are used, then the id of the object *must* be
        included.

        Args:
            objects (list): A list of dicts or a list of Record.
        """

        if not isinstance(objects, list):
            raise ValueError("objects must be a list[dict()|Record] not " + str(type(objects)))

        bulk_data = []
        for o in objects:
            try:
                if isinstance(o, dict):
                    bulk_data.append(o)
                elif isinstance(o, Record):
                    if not hasattr(o, "id"):
                        raise ValueError("'Record' object has no attribute 'id'")
                    updates = o.updates()
                    if updates:
                        updates["id"] = o.id
                        bulk_data.append(updates)
                else:
                    raise ValueError("Invalid object type: " + str(type(o)))
            except ValueError as exc:
                raise ValueError("Unexpected value in object list") from exc

        req = Request(
            base=self.url,
            token=self.api.token,
            http_session=self.api.http_session,
            api_version=self.api.api_version,
        ).patch(bulk_data)
        return response_loader(req, self.return_obj, self)

    def delete(self, objects):
        """Bulk deletes objects on an endpoint.

        Allows for batch deletion of multiple objects from
        a single endpoint.

        Args:
            objects (list): A list of either IDs or Records to delete.

        Returns:
            bool: True if bulk DELETE operation was successful.

        Examples:
            Deleting all `devices`:
            >>> pynautobot.dcim.devices.delete(pynautobot.dcim.devices.all())

            Use bulk deletion by passing a list of IDs:
            >>> pynautobot.dcim.devices.delete([
            ...     "db8770c4-61e5-4999-8372-e7fa576a4f65",
            ...     "e9b5f2e0-4f20-41ad-9179-90a4987f743e"
            ... ])

            Use bulk deletion to delete objects e.g. when filtering
            on a `custom_field`:
            >>> pynautobot.dcim.devices.delete([
            ...     d for d in pynautobot.dcim.devices.all()
            ...     if d.custom_fields.get("field", False)
            ... ])
        """

        ids = []
        if not isinstance(objects, list):
            raise ValueError("objects must be a list[str(id)|Record] not " + str(type(objects)))
        for o in objects:
            try:
                if isinstance(o, str):
                    if UUID(o):
                        ids.append(o)
                elif isinstance(o, Record):
                    if not hasattr(o, "id"):
                        raise ValueError("'Record' object has no attribute 'id'")
                    ids.append(o.id)
                else:
                    raise ValueError("Invalid object type: " + str(type(o)))
            except ValueError as exc:
                raise ValueError("Unexpected value in object list") from exc

        req = Request(
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
            api_version=self.api.api_version,
        )

        return req.delete(data=[{"id": id} for id in ids])

    def choices(self, api_version=None):
        """Returns all choices from the endpoint.

        The returned dict is also saved in the endpoint object (in
        ``_choices`` attribute) so that later calls will return the same data
        without recurring requests to Nautobot. When using ``.choices()`` in
        long-running applications, consider restarting them whenever Nautobot is
        upgraded, to prevent using stale choices data.

        Args:
            api_version (str, optional): Override default or globally-set
                Nautobot REST API version for this single request.

        Returns:
            dict: Dict containing the available choices.

        Example (from Nautobot 2.8.x):
            >>> from pprint import pprint
            >>> pprint(nb.ipam.ip_addresses.choices())
            {'role': [{'display_name': 'Loopback', 'value': 'loopback'},
                    {'display_name': 'Secondary', 'value': 'secondary'},
                    {'display_name': 'Anycast', 'value': 'anycast'},
                    {'display_name': 'VIP', 'value': 'vip'},
                    {'display_name': 'VRRP', 'value': 'vrrp'},
                    {'display_name': 'HSRP', 'value': 'hsrp'},
                    {'display_name': 'GLBP', 'value': 'glbp'},
                    {'display_name': 'CARP', 'value': 'carp'}],
            'status': [{'display_name': 'Active', 'value': 'active'},
                        {'display_name': 'Reserved', 'value': 'reserved'},
                        {'display_name': 'Deprecated', 'value': 'deprecated'},
                        {'display_name': 'DHCP', 'value': 'dhcp'}]}
        """
        if self._choices:
            return self._choices

        api_version = api_version or self.api.api_version

        req = Request(
            base=self.url,
            token=self.api.token,
            http_session=self.api.http_session,
            api_version=api_version,
        ).options()
        try:
            post_data = req["schema"]["properties"]
        except (KeyError, TypeError):
            raise ValueError(f"Unexpected format in the OPTIONS response at {self.url}")
        self._choices = {
            prop: [{"value": x, "display": y} for x, y in zip(post_data[prop]["enum"], post_data[prop]["enumNames"])]
            for prop in post_data
            if "enum" in post_data[prop]
        }
        return self._choices

    def count(self, *args, api_version=None, **kwargs):
        """Returns the count of objects in a query.

        Takes named arguments that match the usable filters on a
        given endpoint. If an argument is passed then it's used as a
        freeform search argument if the endpoint supports it. If no
        arguments are passed the count for all objects on an endpoint
        are returned.

        Args:
            *args (str, optional): Freeform search string that's
                accepted on given endpoint.
            **kwargs (str, optional): Any search argument the
                endpoint accepts can be added as a keyword arg.
            api_version (str, optional): Override default or globally-set
                Nautobot REST API version for this single request.

        Returns:
            int: Integer with count of objects returned by query.

        Examples:
            To return a count of objects matching a named argument filter.
            >>> nb.dcim.devices.count(site='tst1')
            5827

            To return a count of objects on an entire endpoint.
            >>> nb.dcim.devices.count()
            87382
        """

        if args:
            kwargs.update({"q": args[0]})

        if any(i in RESERVED_KWARGS for i in kwargs):
            raise ValueError("A reserved {} kwarg was passed. Please remove it " "try again.".format(RESERVED_KWARGS))

        api_version = api_version or self.api.api_version

        ret = Request(
            filters=kwargs, base=self.url, token=self.token, http_session=self.api.http_session, api_version=api_version
        )

        return ret.get_count()


class DetailEndpoint(object):
    """Enables read/write Operations on detail endpoints.

    Endpoints like ``available-ips`` that are detail routes off
    traditional endpoints are handled with this class.
    """

    def __init__(self, parent_obj, name, custom_return=None):
        self.parent_obj = parent_obj
        self.custom_return = custom_return
        self.url = "{}/{}/{}/".format(parent_obj.endpoint.url, parent_obj.id, name)

        self.request_kwargs = dict(
            base=self.url,
            token=parent_obj.api.token,
            http_session=parent_obj.api.http_session,
        )

    def list(self, api_version=None, **kwargs):
        """The view operation for a detail endpoint.

        Returns the response from Nautobot for a detail endpoint.

        Args:
            api_version (str, optional): Override default or globally set Nautobot REST API version for this single request.
            **kwargs: Key/value pairs that get converted into URL parameters when passed to the endpoint.
                E.g. ``.list(method='get_facts')`` would be converted to ``.../?method=get_facts``.

        Returns:
            Union[Dict, List[Dict]]: A dictionary or list of dictionaries retrieved from Nautobot.
        """

        api_version = api_version or self.parent_obj.api.api_version

        req = Request(api_version=api_version, **self.request_kwargs).get(add_params=kwargs)

        if self.custom_return:
            return response_loader(req, self.custom_return, self.parent_obj.endpoint)
        return req

    def create(self, data=None, api_version=None):
        """The write operation for a detail endpoint.

        Creates objects on a detail endpoint in Nautobot.

        Args:
            data (dict or list, optional): A dictionary containing the
                key/value pairs of the items you're creating on the parent
                object. Defaults to an empty dict, which will create a single
                item with default values.
            api_version (str, optional): Override default or globally set
                Nautobot REST API version for this single request.

        Returns:
            Union[Dict, List[Dict]]: A dictionary or list of dictionaries representing
                the items created in Nautobot.
        """
        data = data or {}
        api_version = api_version or self.parent_obj.api.api_version

        req = Request(api_version=api_version, **self.request_kwargs).post(data)
        if self.custom_return:
            return response_loader(req, self.custom_return, self.parent_obj.endpoint)
        return req


class RODetailEndpoint(DetailEndpoint):
    def create(self, data=None, api_version=None):
        raise NotImplementedError("Writes are not supported for this endpoint.")


class JobsEndpoint(Endpoint):
    """Extend Endpoint class to support run method only for jobs."""

    def run(self, *args, api_version=None, **kwargs):
        """Runs a job based on the class_path provided to the job.

        Takes a kwarg of `class_path` or `job_id` to specify the job that should be run.

        Args:
            *args (str, optional): Freeform search string that's
                accepted on given endpoint.
            **kwargs (str, optional): Any search argument the
                endpoint accepts can be added as a keyword arg.
            api_version (str, optional): Override default or globally-set
                Nautobot REST API version for this single request.

        Returns:
            obj: Job details: job_result object uuid found at `obj.result.id`.

        Examples:
            To run a job for verifying hostnames:
            >>> nb.extras.jobs.run(
                    class_path="local/data_quality/VerifyHostnames",
                    data={"hostname_regex": ".*"},
                    commit=True,
                )
        """
        api_version = api_version or self.api.api_version or self.api.version

        # Check Nautobot Version as API version can be `None`.  Job run endpoints changed in 1.3.
        if float(api_version) < 1.3:
            if not kwargs.get("class_path"):
                raise ValueError(
                    'Keyword Argument "class_path" is required to run a job in Nautobot APIv1.2 and older.'
                )
            job_run_url = f"{self.url}/{kwargs['class_path']}/run/"
        else:
            if not kwargs.get("job_id"):
                raise ValueError('Keyword Argument "job_id" is required to run a job in Nautobot APIv1.3 and newer.')
            job_run_url = f"{self.url}/{kwargs['job_id']}/run/"

        req = Request(
            base=job_run_url,
            token=self.token,
            http_session=self.api.http_session,
            api_version=api_version,
        ).post(args[0] if args else kwargs)

        return response_loader(req, self.return_obj, self)



# (c) 2017 DigitalOcean
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by NetworktoCode, LLC.

from pynautobot.core.endpoint import JobsEndpoint, DetailEndpoint
from pynautobot.core.response import JsonField, Record


class ConfigContexts(Record):
    data = JsonField


class ObjectChanges(Record):
    object_data = JsonField

    def __str__(self):
        return self.request_id


class CustomFieldChoices(Record):
    def __str__(self):
        return self.value


class JobResults(Record):
    data = JsonField


class Jobs(Record):
    def run(self, **kwargs):
        """Run a job from within a job instance."""
        return JobsEndpoint(self.api, self.api.extras, "jobs").run(class_path=self.id, **kwargs)


class DynamicGroups(Record):
    def __str__(self):
        parent_record_string = super().__str__()
        return parent_record_string or str(self.id)

    @property
    def members(self):
        """Represents the ``members`` detail endpoint.

        Returns a list of DetailEndpoint objects that are
        related to the dynamic group

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        Dynamic group of devices:

        >>> group = nb.extras.dynamic_groups.get("device-group")
        >>> group.members.list()
        [<pynautobot.models.extras.DynamicGroups ('testswitch') at 0x7efee4037e80>...]

        Dynamic group of IPs:

        >>> group = nb.extras.dynamic_groups.get("ip-group")
        >>> group.members.list()
        [<pynautobot.models.extras.DynamicGroups ('192.168.10.200/32') at 0x7f3e6a980040>...]
        """
        return DetailEndpoint(self, "members", custom_return=DynamicGroups)
